pub mod api;
pub mod config;

use axum::{
    extract::State,
    http::{
        header::{self, *},
        Method, StatusCode, Uri,
    },
    response::{IntoResponse, Response},
    routing::get,
    Router,
};
use std::net::SocketAddr;
use std::sync::Arc;
use tower_http::trace::{self, TraceLayer};
use tracing::Level;

use crate::config::Config;

pub struct Context {
    pub config: Config,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    tracing_subscriber::fmt()
        .with_target(false)
        .compact()
        .init();

    let config = match Config::open(Config::data_dir()?.join("config.toml").as_path()) {
        Ok(config) => {
            println!("Config loaded: {:?}", config);
            config
        }
        Err(err) => {
            eprintln!("{}", err);
            Config::new()
        }
    };

    let state = Arc::new(Context {
        config: config.clone(),
    });

    let app = Router::new()
        .route("/", get(frontend_handler))
        .with_state(state.clone())
        .nest("/api", api::routes(state))
        .route("/assets/*file", get(assets))
        .layer(
            TraceLayer::new_for_http()
                .make_span_with(trace::DefaultMakeSpan::new().level(Level::INFO))
                .on_response(trace::DefaultOnResponse::new().level(Level::INFO)),
        );

    let address: SocketAddr =
        format!("{}:{}", config.server.address, config.server.port).parse()?;

    let listener = tokio::net::TcpListener::bind(&address).await?;

    println!("Listening on http://{}", address);

    axum::serve(listener, app.into_make_service()).await?;

    Ok(())
}

async fn assets(uri: Uri) -> Result<impl IntoResponse, ResourceError> {
    let path = uri.path().trim_start_matches("/assets/").to_string();

    match elnafo_radio_frontend::Assets::get(&path) {
        Some(content) => {
            let mime = mime_guess::from_path(path).first_or_octet_stream();
            Ok(([(header::CONTENT_TYPE, mime.as_ref())], content.data).into_response())
        }
        None => Err(ResourceError::NotFound),
    }
}

async fn frontend_handler(State(state): State<Arc<Context>>, _: Uri) -> impl IntoResponse {
    use elnafo_radio_frontend::Meta;

    elnafo_radio_frontend::BaseTemplate {
        view: String::from("app"),
        title: state.config.base.title.clone(),
        meta: if let Some(meta) = &state.config.base.meta {
            Some(
                meta.iter()
                    .map(|kv| Meta {
                        name: kv.0.clone(),
                        content: kv.1.clone(),
                    })
                    .collect(),
            )
        } else {
            None
        },
    }
}

#[derive(Debug)]
pub enum ResourceError {
    NotFound,
    NotExists,
    BadFormat,
    BadContent,
}

impl std::error::Error for ResourceError {}

impl std::fmt::Display for ResourceError {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        match self {
            Self::NotFound | Self::NotExists => write!(f, "Resource was not found"),
            Self::BadFormat => write!(f, "Cannot determine file format"),
            Self::BadContent => write!(f, "Failed to read a file content"),
        }
    }
}

impl IntoResponse for ResourceError {
    fn into_response(self) -> Response {
        let status = match self {
            Self::NotFound => StatusCode::NOT_FOUND,
            Self::NotExists => StatusCode::NO_CONTENT,
            Self::BadFormat | Self::BadContent => StatusCode::INTERNAL_SERVER_ERROR,
        };

        (status, format!("{}", self)).into_response()
    }
}
